package main

import (
	"context"
	"math/rand"
	"sync"
	"sync/atomic"
	"testing"
	"time"
)

const max = 5
const goroutines = 1000

func storeSeen(mu *sync.Mutex, storage map[int32]int, value int32) {
	mu.Lock()
	if v, ok := storage[value]; ok {
		storage[value] = v + 1
	} else {
		storage[value] = 1
	}
	mu.Unlock()
}

func dumpSeen(t *testing.T, storage map[int32]int) {
	t.Logf("0 and %d should have 0 occurences, everything else should have the same(+/- 1)", max+1)
	for i := int32(0); i <= max+1; i++ {
		count := storage[i]
		t.Logf("%d = %-3d", i, count)
	}
}

// 動作の確認
func TestCounter(t *testing.T) {
	delay := time.Duration(rand.Intn(10)) * time.Millisecond

	t.Run("Add/CAS", func(t *testing.T) {
		ctx, cancel := context.WithTimeout(context.Background(), 2*time.Second)
		defer cancel()

		var counter atomic.Int32
		counter.Store(0)

		var wg sync.WaitGroup
		wg.Add(goroutines)
		seen := make(map[int32]int)
		var mu sync.Mutex
		for i := 0; i < goroutines; i++ {
			go func(ctx context.Context, wg *sync.WaitGroup, i int) {
				defer wg.Done()
				for {
					select {
					case <-ctx.Done():
						return
					default:
					}

					current := counter.Add(1)
					if current >= max {
						counter.CompareAndSwap(current, 0)
					}

					storeSeen(&mu, seen, current)

					// t.Logf("goroutine %d: %d", i, current)
					time.Sleep(delay)
				}
			}(ctx, &wg, i)
		}

		wg.Wait()

		dumpSeen(t, seen)
	})

	t.Run("Load/Store", func(t *testing.T) {
		ctx, cancel := context.WithTimeout(context.Background(), 2*time.Second)
		defer cancel()

		var counter atomic.Int32
		counter.Store(1)

		var wg sync.WaitGroup
		wg.Add(goroutines)
		seen := make(map[int32]int)
		var mu sync.Mutex
		for i := 0; i < goroutines; i++ {
			go func(ctx context.Context, wg *sync.WaitGroup, i int) {
				defer wg.Done()
				for {
					select {
					case <-ctx.Done():
						return
					default:
					}

					current := counter.Load()
					if current == max {
						counter.Store(0)
					}
					counter.Add(1)

					storeSeen(&mu, seen, current)

					// t.Logf("goroutine %d: %d", i, current)
					time.Sleep(delay)
				}
			}(ctx, &wg, i)
		}

		wg.Wait()

		dumpSeen(t, seen)
	})

	t.Run("Load/Store", func(t *testing.T) {
		ctx, cancel := context.WithTimeout(context.Background(), 2*time.Second)
		defer cancel()

		var counter int32 = 1

		var wg sync.WaitGroup
		wg.Add(goroutines)
		seen := make(map[int32]int)
		var mu sync.Mutex
		for i := 0; i < goroutines; i++ {
			go func(ctx context.Context, wg *sync.WaitGroup, i int) {
				defer wg.Done()
				for {
					select {
					case <-ctx.Done():
						return
					default:
					}

					mu.Lock()
					current := counter
					if current == max {
						counter = 0
					}
					counter += 1
					mu.Unlock()

					storeSeen(&mu, seen, current)

					// t.Logf("goroutine %d: %d", i, current)
					time.Sleep(delay)
				}
			}(ctx, &wg, i)
		}

		wg.Wait()

		dumpSeen(t, seen)
	})
}

// カウンタに1追加していく操作のベンチマーク
func BenchmarkCounter(b *testing.B) {
	b.Run("nothing", func(b *testing.B) {
		var counter int32
		b.ResetTimer()
		for i := 0; i < b.N; i++ {
			counter++
		}
	})
	b.Run("sync/atomic", func(b *testing.B) {
		var counter atomic.Int32
		b.ResetTimer()
		for i := 0; i < b.N; i++ {
			counter.Add(1)
		}
	})
	b.Run("sync.Mutex", func(b *testing.B) {
		var mu sync.Mutex
		var counter int32
		b.ResetTimer()
		for i := 0; i < b.N; i++ {
			mu.Lock()
			counter++
			mu.Unlock()
		}
	})
}
